from ccinput.tests.testing_utilities import InputTests
from ccinput.exceptions import InvalidParameter, ImpossibleCalculation


class OrcaTests(InputTests):
    def test_sp_SE(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "AM1",
            "charge": "-1",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP AM1
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 1000
        %pal
        nprocs 1
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_HF(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_HF_SMD(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "SMD",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        smd true
        SMDsolvent "chloroform"
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_HF_SMD_lowercase(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "smd",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        smd true
        SMDsolvent "chloroform"
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_HF_SMD18(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "I.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "SMD",
            "solvation_radii": "SMD18",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        I 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        smd true
        SMDsolvent "chloroform"
        radius[53] 2.74
        radius[35] 2.60
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_HF_SMD18_lowercase(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "I.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "smd",
            "solvation_radii": "smd18",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        I 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        smd true
        SMDsolvent "chloroform"
        radius[53] 2.74
        radius[35] 2.60
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_solvation_octanol_smd(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "octanol",
            "solvation_model": "SMD",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        smd true
        SMDsolvent "1-octanol"
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_HF_CPCM(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "CPCM",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G CPCM(chloroform)
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_solvent_synonym(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "CHCL3",
            "solvation_model": "SMD",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        smd true
        SMDsolvent "chloroform"
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_solvation_octanol_cpcm1(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "octanol",
            "solvation_model": "CPCM",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G CPCM(octanol)
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_solvation_octanol_cpcm2(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "1-octanol",
            "solvation_model": "CPCM",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G CPCM(octanol)
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_invalid_solvation(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "ABC",
        }

        with self.assertRaises(Exception):
            self.generate_calculation(**params)

    def test_sp_DFT(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "M06-2X",
            "basis_set": "Def2-SVP",
            "charge": "-1",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP M062X Def2-SVP
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_DFT_specifications(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "M06-2X",
            "basis_set": "Def2-SVP",
            "charge": "-1",
            "specifications": "TightSCF",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP M062X Def2-SVP tightscf
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_DFT_multiple_specifications(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "M06-2X",
            "basis_set": "Def2-SVP",
            "charge": "-1",
            "specifications": "TightSCF GRID6",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP M062X Def2-SVP tightscf grid6
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_DFT_duplicate_specifications(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "M06-2X",
            "basis_set": "Def2-SVP",
            "charge": "-1",
            "specifications": "tightscf TightSCF GRID6",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP M062X Def2-SVP tightscf grid6
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_MP2(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "RI-MP2",
            "basis_set": "cc-pVTZ",
            "charge": "-1",
            "specifications": "cc-pvtz/C",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP RI-MP2 cc-pVTZ cc-pVTZ/C
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_MP2_default_aux(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "RI-MP2",
            "basis_set": "cc-pVTZ",
            "charge": "-1",
            "specifications": "",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP RI-MP2 cc-pVTZ cc-pVTZ/C
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_MP2_override_aux(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "RI-MP2",
            "basis_set": "cc-pVTZ",
            "charge": "-1",
            "specifications": "cc-pVDZ/C",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP RI-MP2 cc-pVTZ cc-pVDZ/C
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_MP2_without_RI(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "MP2",
            "basis_set": "cc-pVTZ",
            "charge": "-1",
            "specifications": "",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP MP2 cc-pVTZ
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_DLPNO_CCSDT(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "DLPNO-CCSD(T)",
            "basis_set": "cc-pVTZ",
            "charge": "-1",
            "specifications": "cc-pVTZ/C",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP DLPNO-CCSD(T) cc-pVTZ cc-pVTZ/C
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_DLPNO_CCSDT_default_aux(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "DLPNO-CCSD(T)",
            "basis_set": "cc-pVTZ",
            "charge": "-1",
            "specifications": "",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP DLPNO-CCSD(T) cc-pVTZ cc-pVTZ/C
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_DLPNO_CCSDT_default_aux_specifications(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "DLPNO-CCSD(T)",
            "basis_set": "cc-pVTZ",
            "charge": "-1",
            "specifications": "tightscf",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP DLPNO-CCSD(T) cc-pVTZ cc-pVTZ/C tightscf
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_DLPNO_CCSDT_default_aux_solvation(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "DLPNO-CCSD(T)",
            "basis_set": "cc-pVTZ",
            "charge": "-1",
            "solvent": "ch2cl2",
            "solvation_model": "CPCM",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP DLPNO-CCSD(T) cc-pVTZ cc-pVTZ/C CPCM(ch2cl2)
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_sp_CCSDT(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "CCSD(T)",
            "basis_set": "cc-pVTZ",
            "charge": "-1",
            "specifications": "",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP CCSD(T) cc-pVTZ
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_opt_SE(self):
        params = {
            "nproc": 8,
            "type": "Geometrical Optimisation",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "AM1",
            "charge": "-1",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT AM1
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 1000
        %pal
        nprocs 1
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_opt_HF(self):
        params = {
            "nproc": 8,
            "type": "Geometrical Optimisation",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_opt_DFT(self):
        params = {
            "nproc": 8,
            "type": "Geometrical Optimisation",
            "file": "Cl.xyz",
            "software": "ORCA",
            "charge": "-1",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_freq_SE(self):
        params = {
            "nproc": 8,
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "AM1",
            "charge": "-1",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !FREQ AM1
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 1000
        %pal
        nprocs 1
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_freq_HF(self):
        params = {
            "nproc": 8,
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !FREQ HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_freq_DFT(self):
        params = {
            "nproc": 8,
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "ORCA",
            "charge": "-1",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !FREQ B3LYP 6-31+G(d,p)
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    # opt mod SE and HF

    def test_scan_bond_DFT(self):
        params = {
            "nproc": 8,
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "constraints": "Scan_9_1.4_10/1_2;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz 0 1
        C         -1.31970       -0.64380        0.00000
        H         -0.96310       -1.65260        0.00000
        H         -0.96310       -0.13940       -0.87370
        H         -2.38970       -0.64380        0.00000
        C         -0.80640        0.08220        1.25740
        H         -1.16150        1.09160        1.25640
        H         -1.16470       -0.42110        2.13110
        O          0.62360        0.07990        1.25870
        H          0.94410        0.53240        2.04240
        *
        %geom Scan
        B 0 1 = 9.00, 1.40, 10
        end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_scan_bond_DFT_auto(self):
        params = {
            "nproc": 8,
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "constraints": "Scan_auto_1.4_10/1_2;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz 0 1
        C         -1.31970       -0.64380        0.00000
        H         -0.96310       -1.65260        0.00000
        H         -0.96310       -0.13940       -0.87370
        H         -2.38970       -0.64380        0.00000
        C         -0.80640        0.08220        1.25740
        H         -1.16150        1.09160        1.25640
        H         -1.16470       -0.42110        2.13110
        O          0.62360        0.07990        1.25870
        H          0.94410        0.53240        2.04240
        *
        %geom Scan
        B 0 1 = 1.07, 1.40, 10
        end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_invalid_opt_mod(self):
        params = {
            "nproc": 8,
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "constraints": "",
        }

        with self.assertRaises(Exception):
            self.generate_calculation(**params)

    def test_no_method(self):
        params = {
            "nproc": 8,
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "",
            "basis_set": "6-31+G(d,p)",
            "constraints": "",
        }

        with self.assertRaises(Exception):
            self.generate_calculation(**params)

    def test_scan_angle_DFT(self):
        params = {
            "nproc": 8,
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "constraints": "Scan_9_90_10/2_1_3;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz 0 1
        C         -1.31970       -0.64380        0.00000
        H         -0.96310       -1.65260        0.00000
        H         -0.96310       -0.13940       -0.87370
        H         -2.38970       -0.64380        0.00000
        C         -0.80640        0.08220        1.25740
        H         -1.16150        1.09160        1.25640
        H         -1.16470       -0.42110        2.13110
        O          0.62360        0.07990        1.25870
        H          0.94410        0.53240        2.04240
        *
        %geom Scan
        A 1 0 2 = 9.00, 90.00, 10
        end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_scan_dihedral_DFT(self):
        params = {
            "nproc": 8,
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "constraints": "Scan_9_1_10/4_1_5_8;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz 0 1
        C         -1.31970       -0.64380        0.00000
        H         -0.96310       -1.65260        0.00000
        H         -0.96310       -0.13940       -0.87370
        H         -2.38970       -0.64380        0.00000
        C         -0.80640        0.08220        1.25740
        H         -1.16150        1.09160        1.25640
        H         -1.16470       -0.42110        2.13110
        O          0.62360        0.07990        1.25870
        H          0.94410        0.53240        2.04240
        *
        %geom Scan
        D 3 0 4 7 = 9.00, 1.00, 10
        end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_scan_no_constraint(self):
        params = {
            "nproc": 8,
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
        }

        with self.assertRaises(InvalidParameter):
            self.generate_calculation(**params)

    def test_freeze_bond_DFT(self):
        params = {
            "nproc": 8,
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "constraints": "Freeze/1_2;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz 0 1
        C         -1.31970       -0.64380        0.00000
        H         -0.96310       -1.65260        0.00000
        H         -0.96310       -0.13940       -0.87370
        H         -2.38970       -0.64380        0.00000
        C         -0.80640        0.08220        1.25740
        H         -1.16150        1.09160        1.25640
        H         -1.16470       -0.42110        2.13110
        O          0.62360        0.07990        1.25870
        H          0.94410        0.53240        2.04240
        *
        %geom Constraints
        { B 0 1 C }
        end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_freeze_angle_DFT(self):
        params = {
            "nproc": 8,
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "constraints": "Freeze/2_1_3;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz 0 1
        C         -1.31970       -0.64380        0.00000
        H         -0.96310       -1.65260        0.00000
        H         -0.96310       -0.13940       -0.87370
        H         -2.38970       -0.64380        0.00000
        C         -0.80640        0.08220        1.25740
        H         -1.16150        1.09160        1.25640
        H         -1.16470       -0.42110        2.13110
        O          0.62360        0.07990        1.25870
        H          0.94410        0.53240        2.04240
        *
        %geom Constraints
        { A 1 0 2 C }
        end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_freeze_dihedral_DFT(self):
        params = {
            "nproc": 8,
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "constraints": "Freeze/4_1_5_8;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz 0 1
        C         -1.31970       -0.64380        0.00000
        H         -0.96310       -1.65260        0.00000
        H         -0.96310       -0.13940       -0.87370
        H         -2.38970       -0.64380        0.00000
        C         -0.80640        0.08220        1.25740
        H         -1.16150        1.09160        1.25640
        H         -1.16470       -0.42110        2.13110
        O          0.62360        0.07990        1.25870
        H          0.94410        0.53240        2.04240
        *
        %geom Constraints
        { D 3 0 4 7 C }
        end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_nmr_DFT(self):
        params = {
            "nproc": 8,
            "type": "NMR Prediction",
            "file": "Cl.xyz",
            "software": "ORCA",
            "charge": "-1",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !NMR B3LYP 6-31+G(d,p)
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_irrelevant_gen_bs(self):
        params = {
            "nproc": 8,
            "type": "NMR Prediction",
            "file": "Cl.xyz",
            "software": "ORCA",
            "charge": "-1",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "custom_basis_sets": "N=Def2-SVP;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !NMR B3LYP 6-31+G(d,p)
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_ts_DFT(self):
        params = {
            "nproc": 8,
            "type": "TS Optimisation",
            "file": "mini_ts.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPTTS B3LYP 6-31+G(d,p)
        *xyz 0 1
        N   1.08764072053386     -0.33994563112543     -0.00972525479568
        H   1.99826836912112      0.05502842705407      0.00651240826058
        H   0.59453997172323     -0.48560162159600      0.83949232123172
        H   0.66998093862168     -0.58930117433261     -0.87511947469677
        *
        %geom
        Calc_Hess true
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_ts_xtb(self):
        params = {
            "nproc": 8,
            "type": "TS Optimisation",
            "file": "mini_ts.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "xtb",
        }

        inp = self.generate_calculation(**params)

        # One cannot calculate the Hessian for use in a TS optimization
        # when using xtb as QM engine.
        REF = """
        !OPTTS xtb
        *xyz 0 1
        N   1.08764072053386     -0.33994563112543     -0.00972525479568
        H   1.99826836912112      0.05502842705407      0.00651240826058
        H   0.59453997172323     -0.48560162159600      0.83949232123172
        H   0.66998093862168     -0.58930117433261     -0.87511947469677
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

        # combination tests

    def test_ts_DFT_custom_bs(self):
        params = {
            "nproc": 8,
            "type": "TS Optimisation",
            "file": "mini_ts.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "custom_basis_sets": "N=Def2-SVP;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPTTS B3LYP 6-31+G(d,p)
        *xyz 0 1
        N   1.08764072053386     -0.33994563112543     -0.00972525479568
        H   1.99826836912112      0.05502842705407      0.00651240826058
        H   0.59453997172323     -0.48560162159600      0.83949232123172
        H   0.66998093862168     -0.58930117433261     -0.87511947469677
        *
        %geom
        Calc_Hess true
        end
        %basis
        NewGTO N "Def2-SVP" end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_ts_DFT_custom_bs_synonym(self):
        params = {
            "nproc": 8,
            "type": "TS Optimisation",
            "file": "mini_ts.xyz",
            "software": "ORCA",
            "charge": "0",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "custom_basis_sets": "N=def2Svp;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPTTS B3LYP 6-31+G(d,p)
        *xyz 0 1
        N   1.08764072053386     -0.33994563112543     -0.00972525479568
        H   1.99826836912112      0.05502842705407      0.00651240826058
        H   0.59453997172323     -0.48560162159600      0.83949232123172
        H   0.66998093862168     -0.58930117433261     -0.87511947469677
        *
        %geom
        Calc_Hess true
        end
        %basis
        NewGTO N "Def2-SVP" end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_opt_DFT_custom_bs_ecp(self):
        params = {
            "nproc": 8,
            "type": "Geometrical Optimisation",
            "file": "Ph2I_cation.xyz",
            "software": "ORCA",
            "charge": "+1",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "custom_basis_sets": "I=Def2-TZVPD;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz 1 1
        C         -3.06870       -2.28540        0.00000
        C         -1.67350       -2.28540        0.00000
        C         -0.97600       -1.07770        0.00000
        C         -1.67360        0.13090       -0.00120
        C         -3.06850        0.13080       -0.00170
        C         -3.76610       -1.07740       -0.00070
        H         -3.61840       -3.23770        0.00040
        H         -1.12400       -3.23790        0.00130
        H          0.12370       -1.07760        0.00060
        H         -1.12340        1.08300       -0.00130
        H         -4.86570       -1.07720       -0.00090
        I         -4.11890        1.94920       -0.00350
        C         -4.64360        2.85690       -1.82310
        C         -3.77180        3.76300       -2.42740
        C         -5.86360        2.55380       -2.42750
        C         -4.12020        4.36650       -3.63560
        H         -2.81040        4.00240       -1.95030
        C         -6.21180        3.15650       -3.63650
        H         -6.55070        1.83950       -1.95140
        C         -5.34050        4.06290       -4.24060
        H         -3.43340        5.08120       -4.11170
        H         -7.17360        2.91710       -4.11310
        H         -5.61500        4.53870       -5.19320
        *
        %basis
        NewGTO I "Def2-TZVPD" end
        NewECP I "def2-ECP" end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_opt_DFT_custom_bs_ecp_synonym(self):
        params = {
            "nproc": 8,
            "type": "Geometrical Optimisation",
            "file": "Ph2I_cation.xyz",
            "software": "ORCA",
            "charge": "+1",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "custom_basis_sets": "I=Def2 tzvpd;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz 1 1
        C         -3.06870       -2.28540        0.00000
        C         -1.67350       -2.28540        0.00000
        C         -0.97600       -1.07770        0.00000
        C         -1.67360        0.13090       -0.00120
        C         -3.06850        0.13080       -0.00170
        C         -3.76610       -1.07740       -0.00070
        H         -3.61840       -3.23770        0.00040
        H         -1.12400       -3.23790        0.00130
        H          0.12370       -1.07760        0.00060
        H         -1.12340        1.08300       -0.00130
        H         -4.86570       -1.07720       -0.00090
        I         -4.11890        1.94920       -0.00350
        C         -4.64360        2.85690       -1.82310
        C         -3.77180        3.76300       -2.42740
        C         -5.86360        2.55380       -2.42750
        C         -4.12020        4.36650       -3.63560
        H         -2.81040        4.00240       -1.95030
        C         -6.21180        3.15650       -3.63650
        H         -6.55070        1.83950       -1.95140
        C         -5.34050        4.06290       -4.24060
        H         -3.43340        5.08120       -4.11170
        H         -7.17360        2.91710       -4.11310
        H         -5.61500        4.53870       -5.19320
        *
        %basis
        NewGTO I "Def2-TZVPD" end
        NewECP I "def2-ECP" end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_opt_DFT_custom_bs_explicit(self):
        params = {
            "nproc": 8,
            "type": "Geometrical Optimisation",
            "file": "Ph2I_cation.xyz",
            "software": "ORCA",
            "charge": "+1",
            "method": "B3LYP",
            "basis_set": "6-31+G(d,p)",
            "custom_basis_sets": "I=psbkjc;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT B3LYP 6-31+G(d,p)
        *xyz 1 1
        C         -3.06870       -2.28540        0.00000
        C         -1.67350       -2.28540        0.00000
        C         -0.97600       -1.07770        0.00000
        C         -1.67360        0.13090       -0.00120
        C         -3.06850        0.13080       -0.00170
        C         -3.76610       -1.07740       -0.00070
        H         -3.61840       -3.23770        0.00040
        H         -1.12400       -3.23790        0.00130
        H          0.12370       -1.07760        0.00060
        H         -1.12340        1.08300       -0.00130
        H         -4.86570       -1.07720       -0.00090
        I         -4.11890        1.94920       -0.00350
        C         -4.64360        2.85690       -1.82310
        C         -3.77180        3.76300       -2.42740
        C         -5.86360        2.55380       -2.42750
        C         -4.12020        4.36650       -3.63560
        H         -2.81040        4.00240       -1.95030
        C         -6.21180        3.15650       -3.63650
        H         -6.55070        1.83950       -1.95140
        C         -5.34050        4.06290       -4.24060
        H         -3.43340        5.08120       -4.11170
        H         -7.17360        2.91710       -4.11310
        H         -5.61500        4.53870       -5.19320
        *
        %basis
        newgto I
        S   4
        1         2.6250000              0.07366000
        2         1.0140000             -0.83687000
        3         0.5009000              0.65624700
        4         0.2023000              0.90074400
        S   2
        1         0.2023000              1.00000000
        2         0.0780000              1.00000000
        S   2
        1         0.0780000              1.00000000
        2         0.0300741              1.00000000
        P   4
        1         2.6250000             -0.00888000
        2         1.0140000             -0.25735100
        3         0.5009000              0.45536800
        4         0.2023000              0.76010700
        P   2
        1         0.2023000              1.00000000
        2         0.0780000              1.00000000
        P   2
        1         0.0780000              1.00000000
        2         0.0300741              1.00000000
        D   2
        1         0.5009000              0.36111300
        2         0.2023000              0.93252210
        D   2
        1         0.0780000              0.93733150
        2         0.0300741              0.34843880
        end

        NewECP I
        N_core 46
        lmax f
        s 3
        1      4.1307100     12.1112300 0
        2      1.3337500    -41.0920600 2
        3      1.4912100     70.7376100 2
        p 3
        1      3.0469200     10.5927100 0
        2      1.0633900    -46.0227300 2
        3      1.1440500     65.0504700 2
        d 2
        1      3.9306300      9.7308900 0
        2      1.0692000     13.9888000 2
        f 2
        1      0.9762800     -3.6963900 1
        2      4.3334300    -14.0030500 1
        end
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_NEB(self):
        params = {
            "nproc": 8,
            "type": "Minimum Energy Path",
            "file": "elimination_substrate.xyz",
            "auxiliary_file": "elimination_product.xyz",
            "software": "orca",
            "charge": -1,
            "method": "gfn2-xtb",
        }

        inp = self.generate_calculation(**params)

        REF = """!NEB xtb2
        *xyz -1 1
        C         -0.74277        0.14309        0.12635
        C          0.71308       -0.12855       -0.16358
        Cl         0.90703       -0.47793       -1.61303
        H         -0.84928        0.38704        1.20767
        H         -1.36298       -0.72675       -0.06978
        H         -1.11617        0.99405       -0.43583
        H          1.06397       -0.95639        0.44985
        H          1.30839        0.75217        0.07028
        O         -0.91651        0.74066        3.00993
        H         -1.82448        0.94856        3.28105
        *
        %neb
        product "calc2.xyz"
        nimages 8
        end
        %MaxCore 125
        %pal
        nprocs 8
        end

        """
        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_NEB2(self):
        params = {
            "nproc": 8,
            "type": "Minimum Energy Path",
            "file": "elimination_substrate.xyz",
            "auxiliary_file": "elimination_product.xyz",
            "software": "ORCA",
            "specifications": "--nimages 12",
            "charge": -1,
            "method": "gfn2-xtb",
        }

        inp = self.generate_calculation(**params)

        REF = """!NEB xtb2
        *xyz -1 1
        C         -0.74277        0.14309        0.12635
        C          0.71308       -0.12855       -0.16358
        Cl         0.90703       -0.47793       -1.61303
        H         -0.84928        0.38704        1.20767
        H         -1.36298       -0.72675       -0.06978
        H         -1.11617        0.99405       -0.43583
        H          1.06397       -0.95639        0.44985
        H          1.30839        0.75217        0.07028
        O         -0.91651        0.74066        3.00993
        H         -1.82448        0.94856        3.28105
        *
        %neb
        product "calc2.xyz"
        nimages 12
        end
        %MaxCore 125
        %pal
        nprocs 8
        end

        """
        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_NEB_aux_name(self):
        params = {
            "nproc": 8,
            "type": "Minimum Energy Path",
            "file": "elimination_substrate.xyz",
            "auxiliary_file": "elimination_product.xyz",
            "software": "ORCA",
            "specifications": "--nimages 12",
            "charge": -1,
            "method": "gfn2-xtb",
            "aux_name": "product",
        }

        inp = self.generate_calculation(**params)

        REF = """!NEB xtb2
        *xyz -1 1
        C         -0.74277        0.14309        0.12635
        C          0.71308       -0.12855       -0.16358
        Cl         0.90703       -0.47793       -1.61303
        H         -0.84928        0.38704        1.20767
        H         -1.36298       -0.72675       -0.06978
        H         -1.11617        0.99405       -0.43583
        H          1.06397       -0.95639        0.44985
        H          1.30839        0.75217        0.07028
        O         -0.91651        0.74066        3.00993
        H         -1.82448        0.94856        3.28105
        *
        %neb
        product "product.xyz"
        nimages 12
        end
        %MaxCore 125
        %pal
        nprocs 8
        end

        """
        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_hirshfeld_pop(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "M06-2X",
            "basis_set": "Def2-SVP",
            "charge": "-1",
            "specifications": "--phirshfeld",
        }

        inp = self.generate_calculation(**params)
        REF = """
        !SP M062X Def2-SVP
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %output
        Print[ P_Hirshfeld] 1
        end
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_mo(self):
        params = {
            "nproc": 8,
            "type": "MO Calculation",
            "file": "Ph2I_cation.xyz",
            "software": "ORCA",
            "charge": "+1",
            "method": "B3LYP",
            "basis_set": "def2tzvp",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP B3LYP Def2-TZVP
        *xyz 1 1
        C         -3.06870       -2.28540        0.00000
        C         -1.67350       -2.28540        0.00000
        C         -0.97600       -1.07770        0.00000
        C         -1.67360        0.13090       -0.00120
        C         -3.06850        0.13080       -0.00170
        C         -3.76610       -1.07740       -0.00070
        H         -3.61840       -3.23770        0.00040
        H         -1.12400       -3.23790        0.00130
        H          0.12370       -1.07760        0.00060
        H         -1.12340        1.08300       -0.00130
        H         -4.86570       -1.07720       -0.00090
        I         -4.11890        1.94920       -0.00350
        C         -4.64360        2.85690       -1.82310
        C         -3.77180        3.76300       -2.42740
        C         -5.86360        2.55380       -2.42750
        C         -4.12020        4.36650       -3.63560
        H         -2.81040        4.00240       -1.95030
        C         -6.21180        3.15650       -3.63650
        H         -6.55070        1.83950       -1.95140
        C         -5.34050        4.06290       -4.24060
        H         -3.43340        5.08120       -4.11170
        H         -7.17360        2.91710       -4.11310
        H         -5.61500        4.53870       -5.19320
        *
        %plots
        dim1 45
        dim2 45
        dim3 45
        min1 0
        max1 0
        min2 0
        max2 0
        min3 0
        max3 0
        Format Gaussian_Cube
        MO("in-HOMO.cube",66,0);
        MO("in-LUMO.cube",67,0);
        MO("in-LUMOA.cube",68,0);
        MO("in-LUMOB.cube",69,0);
        end

        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_nproc(self):
        params = {
            "nproc": 1,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 1000
        %pal
        nprocs 1
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_opt_freq(self):
        params = {
            "nproc": 8,
            "type": "opt+freq",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "AM1",
            "charge": "-1",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT FREQ AM1
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 1000
        %pal
        nprocs 1
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_d3(self):
        params = {
            "nproc": 8,
            "type": "opt",
            "file": "Cl.xyz",
            "software": "ORCA",
            "basis_set": "Def2TZVP",
            "method": "M062X",
            "charge": "-1",
            "d3": True,
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT M062X Def2-TZVP d3zero
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_d3bj(self):
        params = {
            "nproc": 8,
            "type": "opt",
            "file": "Cl.xyz",
            "software": "ORCA",
            "basis_set": "Def2TZVP",
            "method": "PBE0",
            "charge": "-1",
            "d3bj": True,
        }

        inp = self.generate_calculation(**params)

        REF = """
        !OPT PBE0 Def2-TZVP d3bj
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_d3_d3bj_crash(self):
        params = {
            "nproc": 8,
            "type": "opt",
            "file": "Cl.xyz",
            "software": "ORCA",
            "basis_set": "Def2TZVP",
            "method": "PBE0",
            "charge": "-1",
            "d3": True,
            "d3bj": True,
        }

        with self.assertRaises(InvalidParameter):
            self.generate_calculation(**params)

    def test_SMD_custom_radius(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "SMD",
            "custom_solvation_radii": "Cl=1.00;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        smd true
        SMDsolvent "chloroform"
        radius[17] 1.00
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_SMD_custom_radii(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "SMD",
            "custom_solvation_radii": "Cl=1.00;Br=2.00;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        smd true
        SMDsolvent "chloroform"
        radius[17] 1.00
        radius[35] 2.00
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_SMD_custom_radius_and_SMD18(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "SMD",
            "solvation_radii": "SMD18",
            "custom_solvation_radii": "Cl=1.00;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        smd true
        SMDsolvent "chloroform"
        radius[17] 1.00
        radius[53] 2.74
        radius[35] 2.60
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_SMD_custom_radius_and_SMD18_clash(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "SMD",
            "solvation_radii": "SMD18",
            "custom_solvation_radii": "I=3.00;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        smd true
        SMDsolvent "chloroform"
        radius[53] 3.00
        radius[35] 2.60
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_CPCM_custom_radius(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "CPCM",
            "custom_solvation_radii": "Cl=1.00;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G CPCM(chloroform)
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        radius[17] 1.00
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_CPCM_custom_radii(self):
        params = {
            "nproc": 8,
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
            "solvent": "Chloroform",
            "solvation_model": "CPCM",
            "custom_solvation_radii": "Cl=1.00;Br=2.00;",
        }

        inp = self.generate_calculation(**params)

        REF = """
        !SP HF 3-21G CPCM(chloroform)
        *xyz -1 1
        Cl 0.0 0.0 0.0
        *
        %MaxCore 125
        %pal
        nprocs 8
        end
        %cpcm
        radius[17] 1.00
        radius[35] 2.00
        end
        """

        self.assertTrue(self.is_equivalent(REF, inp.input_file))

    def test_unavailable_calc_type(self):
        params = {
            "nproc": 8,
            "type": "Conformational Search",
            "file": "Cl.xyz",
            "software": "ORCA",
            "method": "HF",
            "basis_set": "3-21G",
            "charge": "-1",
        }

        with self.assertRaises(ImpossibleCalculation):
            self.generate_calculation(**params)



import copy
import numpy as np

from ccinput.utilities import (
    get_method,
    get_solvent,
    get_basis_set,
    clean_xyz,
)
from ccinput.constants import CalcType, ATOMIC_NUMBER, LOWERCASE_ATOMIC_SYMBOLS
from ccinput.exceptions import InvalidParameter, ImpossibleCalculation
from ccinput.packages.orca import OrcaCalculation


class PysisDriver:
    SUPPORTED_PACKAGES = ["xtb", "orca"]
    IGNORED_ORCA_BLOCKS = ["%MaxCore"]

    SUPPORTED_KEYWORDS = {
        CalcType.TS: "tsopt",
        CalcType.MEP: "cos",
    }

    DEFAULT_SECTION_PARAMS = {
        "tsopt": {
            "type": "rsirfo",
            "do_hess": False,
            "hessian_recalc": 5,
        },
        "calc": {
            "pal": 1,
            "charge": 0,
            "mult": 1,
        },
        "geom": {
            "type": "dlc",
            "fn": "calc.xyz",
        },
        "cos": {
            "type": "gs",
            "max_nodes": 15,
            "climb": "True",
            "reparam_every": 2,
            "reparam_every_full": 3,
        },
        "preopt": {
            "max_cycles": 10,
        },
    }

    """
        Keys of self.sections are sections of the input file (e.g., geom, calc, tsopt).
        The corresponding values are dictionaries containing the key-value pairs within that section.

        Specifications/parameters are given through the specifications field in a pseudo-Gaussian format:
        $ [...] --specifications "geom(type=redund) tsopt(do_hess=False)"
    """

    def __init__(self, calc):
        self.calc = calc
        self.command = "pysis calc.inp"

        self.input_file = ""

        if self.calc.type not in self.SUPPORTED_KEYWORDS:
            raise ImpossibleCalculation(
                f"Pysis does not support calculations of type {self.calc.type} or its support is not implemented in ccinput"
            )

        self.sections = {}

        self.add_section("geom")
        self.handle_main_parameters()

        self.handle_specifications()

        self.create_input_file()

    def clean(self, s):
        WHITELIST = set(
            "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ/()=-,. _"
        )
        return "".join([c for c in s if c in WHITELIST])

    def add_option(self, section, opt, val):
        if section not in self.sections:
            self.add_section(section)

        self.sections[section][opt] = val

    def add_options(self, sec, options):
        for opt, val in options:
            self.add_option(sec, opt, val)

    def add_section(self, section):
        if section not in self.sections:
            if section in self.DEFAULT_SECTION_PARAMS:
                self.sections[section] = copy.deepcopy(
                    self.DEFAULT_SECTION_PARAMS[section]
                )
            else:
                self.sections[section] = {}
        else:
            logger.warning(f"Duplicate specifications for section {section} were given")

    def handle_specifications(self):
        s = self.clean(self.calc.parameters.specifications.lower())

        # Duplicate code from ccinput/packages/Gaussian.py
        # TODO: refactor
        if s.count("(") != s.count(")"):
            raise InvalidParameter("Invalid specifications: parenthesis not matching")

        _specifications = ""
        remove = False
        for c in s:
            if c == " " and remove:
                continue
            _specifications += c
            if c == "(":
                remove = True
            elif c == ")":
                remove = False

        for spec in _specifications.split(" "):
            if spec.strip() == "":
                continue
            if spec.find("(") != -1:
                key, options = spec.split("(")
                options = options.replace(")", "")
                for option in options.split(","):
                    if option.find("=") == -1:
                        raise InvalidParameter(
                            f"Invalid specification {option}: no key/value delimiter (equals sign)"
                        )
                    if option.strip() != "":
                        param, val = option.split("=")
                        self.add_option(key, param, val)
            else:
                raise InvalidParameter(
                    f"Invalid specification {spec}: should be in the form 'section(key1=value1, ...)'"
                )

    def handle_main_parameters(self):
        if self.calc.parameters.software not in self.SUPPORTED_PACKAGES:
            raise InvalidParameter(
                f"{self.calc.parameters.software} is not currently supported with the pysis driver"
            )

        self.add_option("calc", "type", self.calc.parameters.software)
        self.add_option("calc", "charge", self.calc.charge)
        self.add_option("calc", "mult", self.calc.multiplicity)
        self.add_option("calc", "pal", self.calc.nproc)

        if self.calc.type == CalcType.TS:
            self.add_section("tsopt")
        elif self.calc.type == CalcType.MEP:
            self.add_section("cos")
            self.add_section("preopt")
            self.add_section("opt")

            self.sections["opt"]["type"] = "string"
            self.sections["opt"]["stop_in_when_full"] = -1
            self.sections["opt"]["align"] = "False"

            self.sections["geom"][
                "fn"
            ] = f"[{self.calc.name}.xyz, {self.calc.aux_name}.xyz]"

        else:
            raise InvalidParameter(
                f"Calculation type {self.calc.type} is not currently implemented for pysis in ccinput"
            )

        if self.calc.parameters.software == "orca":
            self.add_option("calc", "mem", self.calc.mem)

            _calc = copy.deepcopy(self.calc)
            _calc.type = CalcType.SP
            orca_inp = OrcaCalculation(_calc)

            self.add_option(
                "calc",
                "keywords",
                f"{orca_inp.command_line} {orca_inp.additional_commands}".replace(
                    "  ", " "
                ).replace("SP ", ""),
            )

            cleaned_blocks = []
            for block in orca_inp.blocks:
                block_name = block.split("\n")[0].split()[0]
                if block_name not in self.IGNORED_ORCA_BLOCKS:
                    cleaned_blocks.append(block.replace('"', '\\"'))

            cleaned_blocks_lines = "\n".join(cleaned_blocks)
            if cleaned_blocks_lines.strip() != "":
                self.add_option("calc", "blocks", f'"{cleaned_blocks_lines}"')
        else:
            if self.calc.parameters.solvent != "":
                if self.calc.parameters.software == "xtb":
                    if self.calc.parameters.solvation_model not in ["gbsa", "alpb"]:
                        raise InvalidParameter(
                            f"Invalid solvation method for xtb: {self.calc.parameters.solvation_model}"
                        )

                    try:
                        solvent_keyword = get_solvent(
                            self.calc.parameters.solvent, self.calc.parameters.software
                        )
                    except KeyError:
                        raise InvalidParameter("Invalid solvent")

                    self.add_option(
                        "calc", self.calc.parameters.solvation_model, solvent_keyword
                    )
                else:
                    raise InvalidParameter(
                        f"Cannot apply solvation for {self.calc.parameters.software}"
                    )
        """
        For other Gaussian:
            method
            basis set
            custom basis set
            constraints
        """

    def create_input_file(self):
        self.input_file = ""
        for section, content in sorted(self.sections.items(), key=lambda i: i[0]):
            self.input_file += f"{section}:\n"
            for key, val in sorted(content.items(), key=lambda i: i[0]):
                self.input_file += f"    {key}: {val}\n"

    @property
    def output(self):
        return self.input_file



import os
import shlex
from unittest import TestCase

from ccinput.wrapper import gen_obj
from ccinput.wrapper import gen_input, get_input_from_args, get_parser


class InputTests(TestCase):
    def generate_calculation(self, **params):
        params["file"] = os.path.join(
            "/".join(__file__.split("/")[:-1]), "structures/", params["file"]
        )
        return gen_obj(**params)

    def args_cmd_equivalent(self, api_args, cmd_line):
        ref = gen_input(**api_args)

        parser = get_parser()
        args = parser.parse_args(shlex.split(cmd_line))
        default_params = vars(parser.parse_args([]))
        calcs, outputs = get_input_from_args(args, default_params=default_params)
        inp = calcs[0].input_file
        return self.is_equivalent(ref, inp)

    def is_equivalent(self, ref, res):
        ref_lines = [i.strip() for i in ref.strip().split("\n")]
        res_lines = [i.strip() for i in res.strip().split("\n")]

        if len(ref_lines) != len(res_lines):
            print(f"Different number of lines: {len(ref_lines)} and {len(res_lines)}")
            print("----START REFERENCE---")
            print("\n".join(ref_lines))
            print("----END REFERENCE---")
            print("----START RESULT---")
            print("\n".join(res_lines))
            print("----END RESULT---")
            return False

        for line1, line2 in zip(ref_lines, res_lines):
            if line1 != line2:
                if self.xyz_line_equivalent(line1, line2):
                    continue
                print("")
                print("Difference found:")
                print("----REFERENCE---")
                print(line1)
                print()
                print("----RESULT---")
                print(line2)
                print("")
                return False

        return True

    def xyz_line_equivalent(self, line1, line2):
        sline1 = line1.strip().split()
        sline2 = line2.strip().split()

        if len(sline1) != 4 or len(sline2) != 4:
            return False

        if sline1[0] != sline2[0]:
            return False

        for c1, c2 in zip(sline1[1:], sline2[1:]):
            try:
                _c1 = float(c1)
                _c2 = float(c2)
            except ValueError:
                return False
            if abs(_c1 - _c2) > 1e-7:
                return False
        return True

    def struct(self, name):
        return os.path.join(
            "/".join(__file__.split("/")[:-1]), "structures/", name + ".xyz"
        )



from ccinput.tests.testing_utilities import InputTests
from ccinput.packages.xtb import XtbCalculation
from ccinput.exceptions import InvalidParameter, ImpossibleCalculation


class XtbTests(InputTests):
    def test_sp_basic(self):
        params = {
            "type": "Single-Point Energy",
            "file": "ethanol.xyz",
            "software": "xtb",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz"

        self.assertTrue(self.is_equivalent(REF, xtb.command))
        self.assertTrue(self.is_equivalent("", xtb.input_file))

    def test_sp_charge(self):
        params = {
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "xtb",
            "charge": "-1",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb Cl.xyz --chrg -1"

        self.assertTrue(self.is_equivalent(REF, xtb.command))
        self.assertTrue(self.is_equivalent("", xtb.input_file))

    def test_sp_multiplicity(self):
        params = {
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "xtb",
            "multiplicity": "2",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb Cl.xyz --uhf 2"

        self.assertTrue(self.is_equivalent(REF, xtb.command))
        self.assertTrue(self.is_equivalent("", xtb.input_file))

    def test_sp_charge_multiplicity(self):
        params = {
            "type": "Single-Point Energy",
            "file": "Cl.xyz",
            "software": "xtb",
            "multiplicity": "3",
            "charge": "-1",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb Cl.xyz --chrg -1 --uhf 3"

        self.assertTrue(self.is_equivalent(REF, xtb.command))
        self.assertTrue(self.is_equivalent("", xtb.input_file))

    def test_opt_charge(self):
        params = {
            "type": "Geometrical Optimisation",
            "file": "Cl.xyz",
            "software": "xtb",
            "charge": "-1",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb Cl.xyz --opt tight --chrg -1"

        self.assertTrue(self.is_equivalent(REF, xtb.command))
        self.assertTrue(self.is_equivalent("", xtb.input_file))

    def test_freq_charge(self):
        params = {
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "xtb",
            "charge": "-1",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb Cl.xyz --hess --chrg -1"

        self.assertTrue(self.is_equivalent(REF, xtb.command))
        self.assertTrue(self.is_equivalent("", xtb.input_file))

    def test_opt_freq(self):
        params = {
            "type": "Opt+Freq",
            "file": "Cl.xyz",
            "software": "xtb",
            "charge": "-1",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb Cl.xyz --ohess --chrg -1"

        self.assertTrue(self.is_equivalent(REF, xtb.command))
        self.assertTrue(self.is_equivalent("", xtb.input_file))

    def test_solvent(self):
        params = {
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "xtb",
            "solvation_model": "GBSA",
            "solvent": "chcl3",
            "charge": "-1",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb Cl.xyz --hess -g chcl3 --chrg -1"

        self.assertTrue(self.is_equivalent(REF, xtb.command))
        self.assertTrue(self.is_equivalent("", xtb.input_file))

    def test_solvent_ALPB(self):
        params = {
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "xtb",
            "solvent": "chcl3",
            "solvation_model": "ALPB",
            "charge": "-1",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb Cl.xyz --hess --alpb chcl3 --chrg -1"

        self.assertTrue(self.is_equivalent(REF, xtb.command))
        self.assertTrue(self.is_equivalent("", xtb.input_file))

    def test_solvent_invalid_PCM(self):
        params = {
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "xtb",
            "solvent": "chcl3",
            "solvation_model": "PCM",
            "charge": "-1",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_solvent_change_solvation_radii(self):
        params = {
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "xtb",
            "solvent": "chcl3",
            "solvation_model": "ALPB",
            "solvation_radii": "PCM",
            "charge": "-1",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_solvent_custom_solvation_radii(self):
        params = {
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "xtb",
            "solvent": "chcl3",
            "solvation_model": "ALPB",
            "custom_solvation_radii": "I=3.00",
            "charge": "-1",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_solvent_synonym(self):
        params = {
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "xtb",
            "solvent": "chloroform",
            "solvation_model": "GBSA",
            "charge": "-1",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb Cl.xyz --hess -g chcl3 --chrg -1"

        self.assertTrue(self.is_equivalent(REF, xtb.command))
        self.assertTrue(self.is_equivalent("", xtb.input_file))

    def test_solvent_invalid(self):
        params = {
            "type": "Frequency Calculation",
            "file": "Cl.xyz",
            "software": "xtb",
            "solvent": "octanol",
            "solvation_model": "GBSA",
            "charge": "-1",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_scan(self):
        params = {
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Scan_9_1.4_10/1_2;",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --opt tight --input input"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=1.0
        distance: 1, 2, auto
        $scan
        1: 9.00, 1.40, 10
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_scan_auto(self):
        params = {
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Scan_auto_1.4_10/1_2;",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --opt tight --input input"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=1.0
        distance: 1, 2, auto
        $scan
        1: 1.07, 1.40, 10
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_constrained_opt_no_constraint(self):
        params = {
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "xtb",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_freeze(self):
        params = {
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/1_2;",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --opt tight --input input"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=1.0
        distance: 1, 2, auto
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_constraint_overlap(self):
        params = {
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/1_2;Freeze/2_3;",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --opt tight --input input"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=1.0
        distance: 1, 2, auto
        distance: 2, 3, auto
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_freeze_soft(self):
        params = {
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/1_2;",
            "specifications": "--forceconstant 0.1",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --opt tight --input input"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=0.1
        distance: 1, 2, auto
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_duplicate_specifications(self):
        params = {
            "type": "Constrained Optimisation",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/1_2;",
            "specifications": "--forceconstant 0.1 --forceconstant 0.2",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --opt tight --input input"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=0.2
        distance: 1, 2, auto
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_conformational_search(self):
        params = {
            "type": "Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
        }

        xtb = self.generate_calculation(**params)

        REF = "crest ethanol.xyz -rthr 0.6 -ewin 6"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

        self.assertEqual("", xtb.input_file)

    def test_conformational_search_specs(self):
        params = {
            "type": "Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--rthr 0.8 --ewin 8",
        }

        xtb = self.generate_calculation(**params)

        REF = "crest ethanol.xyz -rthr 0.8 -ewin 8"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

        self.assertEqual("", xtb.input_file)

    def test_conformational_search_nci(self):
        params = {
            "type": "Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--rthr 0.8 --nci",
        }

        xtb = self.generate_calculation(**params)

        REF = "crest ethanol.xyz -nci -rthr 0.8 -ewin 6"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

        self.assertEqual("", xtb.input_file)

    def test_constrained_conformational_search1(self):
        params = {
            "type": "Constrained Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/1_2;",
        }

        xtb = self.generate_calculation(**params)

        REF = "crest ethanol.xyz -cinp input -rthr 0.6 -ewin 6"
        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=1.0
        reference=ethanol.xyz
        distance: 1, 2, auto
        atoms: 1-2
        $metadyn
        atoms: 3-9
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_constrained_conformational_search2(self):
        params = {
            "type": "Constrained Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/1_4;Freeze/6_8;",
        }

        xtb = self.generate_calculation(**params)

        REF = "crest ethanol.xyz -cinp input -rthr 0.6 -ewin 6"
        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=1.0
        reference=ethanol.xyz
        distance: 1, 4, auto
        distance: 6, 8, auto
        atoms: 1,4,6,8
        $metadyn
        atoms: 2-3,5,7,9
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_constrained_conformational_search3(self):
        params = {
            "type": "Constrained Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/2_3;",
        }

        xtb = self.generate_calculation(**params)

        REF = "crest ethanol.xyz -cinp input -rthr 0.6 -ewin 6"
        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=1.0
        reference=ethanol.xyz
        distance: 2, 3, auto
        atoms: 2-3
        $metadyn
        atoms: 1,4-9
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_constrained_conformational_search4(self):
        params = {
            "type": "Constrained Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/2_3;",
            "specifications": "--force_constant 2.0",
        }

        xtb = self.generate_calculation(**params)

        REF = "crest ethanol.xyz -cinp input -rthr 0.6 -ewin 6"
        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=2.0
        reference=ethanol.xyz
        distance: 2, 3, auto
        atoms: 2-3
        $metadyn
        atoms: 1,4-9
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_constrained_conformational_search5(self):
        params = {
            "type": "Constrained Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/2_3;Freeze/3_4;",
            "specifications": "--force_constant 2.0",
        }

        xtb = self.generate_calculation(**params)

        REF = "crest ethanol.xyz -cinp input -rthr 0.6 -ewin 6"
        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=2.0
        reference=ethanol.xyz
        distance: 2, 3, auto
        distance: 3, 4, auto
        atoms: 2-4
        $metadyn
        atoms: 1,5-9
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_constrained_conformational_search_named_output(self):
        params = {
            "type": "Constrained Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/1_2;",
            "output": "inp",
        }

        xtb = self.generate_calculation(**params)

        REF = "crest ethanol.xyz -cinp inp -rthr 0.6 -ewin 6"
        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=1.0
        reference=ethanol.xyz
        distance: 1, 2, auto
        atoms: 1-2
        $metadyn
        atoms: 3-9
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_constrained_conformational_search_no_constraint(self):
        params = {
            "type": "Constrained Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--force_constant 2.0",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_constrained_conformational_search_equals(self):
        params = {
            "type": "Constrained Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/2_3;",
            "specifications": "--force_constant=2.0",
        }

        xtb = self.generate_calculation(**params)

        REF = "crest ethanol.xyz -cinp input -rthr 0.6 -ewin 6"
        self.assertTrue(self.is_equivalent(REF, xtb.command))

        INPUT = """$constrain
        force constant=2.0
        reference=ethanol.xyz
        distance: 2, 3, auto
        atoms: 2-3
        $metadyn
        atoms: 1,4-9
        """
        self.assertTrue(self.is_equivalent(INPUT, xtb.input_file))

    def test_invalid_specification(self):
        params = {
            "type": "Constrained Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/2_3;",
            "specifications": "--force 2.0",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_invalid_specification2(self):
        params = {
            "type": "Constrained Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "constraints": "Freeze/2_3;",
            "specifications": "-force_constant 2.0",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_invalid_specification3(self):
        params = {
            "type": "Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "-rthr 0.8 --ewin 8",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_invalid_specification4(self):
        params = {
            "type": "Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--rthr abc",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_invalid_specification_for_type1(self):
        params = {
            "type": "Geometrical Optimisation",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--rthr 0.8",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_invalid_specification_for_type2(self):
        params = {
            "type": "Geometrical Optimisation",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--ewin 8",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_unknown_specification(self):
        params = {
            "type": "Conformational Search",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--abc",
        }

        with self.assertRaises(InvalidParameter):
            xtb = self.generate_calculation(**params)

    def test_gfn0(self):
        params = {
            "type": "Single-Point Energy",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--gfn 0",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --gfn 0"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

    def test_gfn0_2(self):
        params = {
            "type": "Single-Point Energy",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--gfn0",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --gfn 0"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

    def test_gfn1(self):
        params = {
            "type": "Single-Point Energy",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--gfn 0",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --gfn 0"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

    def test_gfn1_method(self):
        params = {
            "type": "Single-Point Energy",
            "method": "gfn0",
            "file": "ethanol.xyz",
            "software": "xtb",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --gfn 0"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

    def test_gfn2(self):
        params = {
            "type": "Single-Point Energy",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--gfn 2",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

    def test_gfnff(self):
        params = {
            "type": "Single-Point Energy",
            "file": "ethanol.xyz",
            "software": "xtb",
            "specifications": "--gfnff",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --gfnff"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

    def test_gfnff_method(self):
        params = {
            "type": "Single-Point Energy",
            "file": "ethanol.xyz",
            "software": "xtb",
            "method": "gfnff",
        }

        xtb = self.generate_calculation(**params)

        REF = "xtb ethanol.xyz --gfnff"

        self.assertTrue(self.is_equivalent(REF, xtb.command))

    def test_unavailable_calc_type(self):
        params = {
            "type": "TS Optimisation",
            "file": "ethanol.xyz",
            "software": "xtb",
        }

        with self.assertRaises(ImpossibleCalculation):
            xtb = self.generate_calculation(**params)
